@@ -8,7 +8,7 @@ from .lib.disk import *
 from .lib.exceptions import *
 from .lib.general import *
 from .lib.hardware import *
-from .lib.installer import __packages__, Installer
+from .lib.installer import __packages__, Installer, accessibility_tools_in_use
 from .lib.locale_helpers import *
 from .lib.luks import *
 from .lib.mirrors import *
@@ -23,14 +23,18 @@ from .lib.user_interaction import *
 
 parser = ArgumentParser()
 
-__version__ = "2.2.0"
+__version__ = "2.3.0"
+storage['__version__'] = __version__
 
 
 def initialize_arguments():
 	config = {}
 	parser.add_argument("--config", nargs="?", help="JSON configuration file or URL")
+	parser.add_argument("--creds", nargs="?", help="JSON credentials configuration file")
 	parser.add_argument("--silent", action="store_true",
-						help="Warning!!! No prompts, ignored if config is not passed")
+						help="WARNING: Disables all prompts for input and confirmation. If no configuration is provided, this is ignored")
+	parser.add_argument("--dry-run", action="store_true",
+						help="Generates a configuration file and then exits instead of performing an installation")
 	parser.add_argument("--script", default="guided", nargs="?", help="Script to run for installation", type=str)
 	args, unknowns = parser.parse_known_args()
 	if args.config is not None:
@@ -45,9 +49,15 @@ def initialize_arguments():
 				with urllib.request.urlopen(urllib.request.Request(args.config, headers={'User-Agent': 'ArchInstall'})) as response:
 					config = json.loads(response.read())
 		except Exception as e:
-			print(e)
+			raise ValueError(f"Could not load --config because: {e}")
+
+		if args.creds is not None:
+			with open(args.creds) as file:
+				config.update(json.load(file))
+
 		# Installation can't be silent if config is not passed
 		config["silent"] = args.silent
+
 	for arg in unknowns:
 		if '--' == arg[:2]:
 			if '=' in arg:
@@ -55,12 +65,26 @@ def initialize_arguments():
 			else:
 				key, val = arg[2:], True
 			config[key] = val
+
 	config["script"] = args.script
+
+	if args.dry_run is not None:
+		config["dry-run"] = args.dry_run
+
 	return config
 
 
 arguments = initialize_arguments()
+storage['arguments'] = arguments
+if arguments.get('debug'):
+	log(f"Warning: --debug mode will write certain credentials to {storage['LOG_PATH']}/{storage['LOG_FILE']}!", fg="red", level=logging.WARNING)
+if arguments.get('mount-point'):
+	storage['MOUNT_POINT'] = arguments['mount-point']
 
+from .lib.plugins import plugins, load_plugin # This initiates the plugin loading ceremony
+
+if arguments.get('plugin', None):
+	load_plugin(arguments['plugin'])
 
 # TODO: Learn the dark arts of argparse... (I summon thee dark spawn of cPython)
 
