@@ -1,13 +1,63 @@
-from .disk import *
-from .hardware import *
+import time
+from typing import Union
+import logging
+import os
+import shutil
+import shlex
+import pathlib
+import subprocess
+import glob
+from .disk import get_partitions_in_use, Partition
+from .general import SysCommand, generate_password
+from .hardware import has_uefi, is_vm, cpu_vendor
 from .locale_helpers import verify_keyboard_layout, verify_x11_keyboard_layout
-from .mirrors import *
+from .disk.helpers import get_mount_info
+from .mirrors import use_mirrors
+from .plugins import plugins
 from .storage import storage
-from .user_interaction import *
+# from .user_interaction import *
+from .output import log
+from .profiles import Profile
+from .disk.btrfs import create_subvolume, mount_subvolume
+from .disk.partition import get_mount_fs_type
+from .exceptions import DiskError, ServiceException, RequirementError, HardwareIncompatibilityError
 
 # Any package that the Installer() is responsible for (optional and the default ones)
 __packages__ = ["base", "base-devel", "linux-firmware", "linux", "linux-lts", "linux-zen", "linux-hardened"]
 
+# Additional packages that are installed if the user is running the Live ISO with accessibility tools enabled
+__accessibility_packages__ = ["brltty", "espeakup", "alsa-utils"]
+
+
+class InstallationFile:
+	def __init__(self, installation, filename, owner, mode="w"):
+		self.installation = installation
+		self.filename = filename
+		self.owner = owner
+		self.mode = mode
+		self.fh = None
+
+	def __enter__(self):
+		self.fh = open(self.filename, self.mode)
+		return self
+
+	def __exit__(self, *args):
+		self.fh.close()
+		self.installation.chown(self.owner, self.filename)
+
+	def write(self, data: Union[str, bytes]):
+		return self.fh.write(data)
+
+	def read(self, *args):
+		return self.fh.read(*args)
+
+	def poll(self, *args):
+		return self.fh.poll(*args)
+
+
+def accessibility_tools_in_use() -> bool:
+	return os.system('systemctl is-active --quiet espeakup.service') == 0
+
 
 class Installer:
 	"""
@@ -39,6 +89,7 @@ class Installer:
 			base_packages = __packages__[:3]
 		if kernels is None:
 			kernels = ['linux']
+		self.kernels = kernels
 		self.target = target
 		self.init_time = time.strftime('%Y-%m-%d_%H-%M-%S')
 		self.milliseconds = int(str(time.time()).split('.')[1])
@@ -52,10 +103,13 @@ class Installer:
 		for kernel in kernels:
 			self.base_packages.append(kernel)
 
+		# If using accessibility tools in the live environment, append those to the packages list
+		if accessibility_tools_in_use():
+			self.base_packages.extend(__accessibility_packages__)
+
 		self.post_base_install = []
 
 		storage['session'] = self
-		self.partitions = get_partitions_in_use(self.target)
 
 		self.MODULES = []
 		self.BINARIES = []
@@ -106,6 +160,10 @@ class Installer:
 			self.sync_log_to_install_medium()
 			return False
 
+	@property
+	def partitions(self):
+		return get_partitions_in_use(self.target)
+
 	def sync_log_to_install_medium(self):
 		# Copy over the install log (if there is one) to the install medium if
 		# at least the base has been strapped in, otherwise we won't have a filesystem/structure to copy to.
@@ -120,6 +178,56 @@ class Installer:
 
 		return True
 
+	def mount_ordered_layout(self, layouts: dict):
+		from .luks import luks2
+
+		mountpoints = {}
+		for blockdevice in layouts:
+			for partition in layouts[blockdevice]['partitions']:
+				mountpoints[partition['mountpoint']] = partition
+
+		for mountpoint in sorted(mountpoints.keys()):
+			partition = mountpoints[mountpoint]
+
+			if partition.get('encrypted', False):
+				loopdev = f"{storage.get('ENC_IDENTIFIER', 'ai')}{pathlib.Path(partition['mountpoint']).name}loop"
+				if not (password := partition.get('!password', None)):
+					raise RequirementError(f"Missing mountpoint {mountpoint} encryption password in layout: {partition}")
+
+				with (luks_handle := luks2(partition['device_instance'], loopdev, password, auto_unmount=False)) as unlocked_device:
+					if partition.get('generate-encryption-key-file'):
+						if not (cryptkey_dir := pathlib.Path(f"{self.target}/etc/cryptsetup-keys.d")).exists():
+							cryptkey_dir.mkdir(parents=True)
+
+						# Once we store the key as ../xyzloop.key systemd-cryptsetup can automatically load this key
+						# if we name the device to "xyzloop".
+						encryption_key_path = f"/etc/cryptsetup-keys.d/{pathlib.Path(partition['mountpoint']).name}loop.key"
+						with open(f"{self.target}{encryption_key_path}", "w") as keyfile:
+							keyfile.write(generate_password(length=512))
+
+						os.chmod(f"{self.target}{encryption_key_path}", 0o400)
+
+						luks_handle.add_key(pathlib.Path(f"{self.target}{encryption_key_path}"), password=password)
+						luks_handle.crypttab(self, encryption_key_path, options=["luks", "key-slot=1"])
+
+					log(f"Mounting {mountpoint} to {self.target}{mountpoint} using {unlocked_device}", level=logging.INFO)
+					unlocked_device.mount(f"{self.target}{mountpoint}")
+
+			else:
+				log(f"Mounting {mountpoint} to {self.target}{mountpoint} using {partition['device_instance']}", level=logging.INFO)
+				partition['device_instance'].mount(f"{self.target}{mountpoint}")
+
+			time.sleep(1)
+			try:
+				get_mount_info(f"{self.target}{mountpoint}", traverse=False)
+			except DiskError:
+				raise DiskError(f"Target {self.target}{mountpoint} never got mounted properly (unable to get mount information using findmnt).")
+
+			if (subvolumes := partition.get('btrfs', {}).get('subvolumes', {})):
+				for name, location in subvolumes.items():
+					create_subvolume(self, location)
+					mount_subvolume(self, location)
+
 	def mount(self, partition, mountpoint, create_mountpoint=True):
 		if create_mountpoint and not os.path.isdir(f'{self.target}{mountpoint}'):
 			os.makedirs(f'{self.target}{mountpoint}')
@@ -132,28 +240,45 @@ class Installer:
 	def pacstrap(self, *packages, **kwargs):
 		if type(packages[0]) in (list, tuple):
 			packages = packages[0]
+
+		for plugin in plugins.values():
+			if hasattr(plugin, 'on_pacstrap'):
+				if (result := plugin.on_pacstrap(packages)):
+					packages = result
+
 		self.log(f'Installing packages: {packages}', level=logging.INFO)
 
 		if (sync_mirrors := SysCommand('/usr/bin/pacman -Syy')).exit_code == 0:
-			if (pacstrap := SysCommand(f'/usr/bin/pacstrap {self.target} {" ".join(packages)}', peak_output=True)).exit_code == 0:
+			if (pacstrap := SysCommand(f'/usr/bin/pacstrap {self.target} {" ".join(packages)} --noconfirm', peak_output=True)).exit_code == 0:
 				return True
 			else:
-				self.log(f'Could not strap in packages: {pacstrap.exit_code}', level=logging.INFO)
+				self.log(f'Could not strap in packages: {pacstrap}', level=logging.ERROR, fg="red")
+				self.log(f'Could not strap in packages: {pacstrap.exit_code}', level=logging.ERROR, fg="red")
+				raise RequirementError("Pacstrap failed. See /var/log/archinstall/install.log or above message for error details.")
 		else:
 			self.log(f'Could not sync mirrors: {sync_mirrors.exit_code}', level=logging.INFO)
 
 	def set_mirrors(self, mirrors):
+		for plugin in plugins.values():
+			if hasattr(plugin, 'on_mirrors'):
+				if result := plugin.on_mirrors(mirrors):
+					mirrors = result
+
 		return use_mirrors(mirrors, destination=f'{self.target}/etc/pacman.d/mirrorlist')
 
 	def genfstab(self, flags='-pU'):
 		self.log(f"Updating {self.target}/etc/fstab", level=logging.INFO)
 
 		with open(f"{self.target}/etc/fstab", 'a') as fstab_fh:
-			fstab_fh.write(SysCommand(f'/usr/bin/genfstab {flags} {self.target}').decode())
+			fstab_fh.write((fstab := SysCommand(f'/usr/bin/genfstab {flags} {self.target}')).decode())
 
 		if not os.path.isfile(f'{self.target}/etc/fstab'):
 			raise RequirementError(f'Could not generate fstab, strapping in packages most likely failed (disk out of space?)\n{fstab}')
 
+		for plugin in plugins.values():
+			if hasattr(plugin, 'on_genfstab'):
+				plugin.on_genfstab(self)
+
 		return True
 
 	def set_hostname(self, hostname: str, *args, **kwargs):
@@ -177,6 +302,11 @@ class Installer:
 		if not len(zone):
 			return True  # Redundant
 
+		for plugin in plugins.values():
+			if hasattr(plugin, 'on_timezone'):
+				if result := plugin.on_timezone(zone):
+					zone = result
+
 		if (pathlib.Path("/usr") / "share" / "zoneinfo" / zone).exists():
 			(pathlib.Path(self.target) / "etc" / "localtime").unlink(missing_ok=True)
 			SysCommand(f'/usr/bin/arch-chroot {self.target} ln -s /usr/share/zoneinfo/{zone} /etc/localtime')
@@ -189,10 +319,25 @@ class Installer:
 			)
 
 	def activate_ntp(self):
-		self.log('Installing and activating NTP.', level=logging.INFO)
-		if self.pacstrap('ntp'):
-			if self.enable_service('ntpd'):
-				return True
+		log(f"activate_ntp() is deprecated, use activate_time_syncronization()", fg="yellow", level=logging.INFO)
+		self.activate_time_syncronization()
+
+	def activate_time_syncronization(self):
+		self.log('Activating systemd-timesyncd for time synchronization using Arch Linux and ntp.org NTP servers.', level=logging.INFO)
+		self.enable_service('systemd-timesyncd')
+
+		with open(f"{self.target}/etc/systemd/timesyncd.conf", "w") as fh:
+			fh.write("[Time]\n")
+			fh.write("NTP=0.arch.pool.ntp.org 1.arch.pool.ntp.org 2.arch.pool.ntp.org 3.arch.pool.ntp.org\n")
+			fh.write("FallbackNTP=0.pool.ntp.org 1.pool.ntp.org 0.fr.pool.ntp.org\n")
+
+		from .systemd import Boot
+		with Boot(self) as session:
+			session.SysCommand(["timedatectl", "set-ntp", 'true'])
+
+	def enable_espeakup(self):
+		self.log('Enabling espeakup.service for speech synthesis (accessibility).', level=logging.INFO)
+		self.enable_service('espeakup')
 
 	def enable_service(self, *services):
 		for service in services:
@@ -200,12 +345,16 @@ class Installer:
 			if (output := self.arch_chroot(f'systemctl enable {service}')).exit_code != 0:
 				raise ServiceException(f"Unable to start service {service}: {output}")
 
+			for plugin in plugins.values():
+				if hasattr(plugin, 'on_service'):
+					plugin.on_service(service)
+
 	def run_command(self, cmd, *args, **kwargs):
 		return SysCommand(f'/usr/bin/arch-chroot {self.target} {cmd}')
 
-	def arch_chroot(self, cmd, *args, **kwargs):
-		if 'runas' in kwargs:
-			cmd = f"su - {kwargs['runas']} -c \"{cmd}\""
+	def arch_chroot(self, cmd, run_as=None):
+		if run_as:
+			cmd = f"su - {run_as} -c {shlex.quote(cmd)}"
 
 		return self.run_command(cmd)
 
@@ -229,6 +378,11 @@ class Installer:
 
 			conf = Networkd(Match={"Name": nic}, Network=network)
 
+		for plugin in plugins.values():
+			if hasattr(plugin, 'on_configure_nic'):
+				if (new_conf := plugin.on_configure_nic(nic, dhcp, ip, gateway, dns)):
+					conf = new_conf
+
 		with open(f"{self.target}/etc/systemd/network/10-{nic}.network", "a") as netconf:
 			netconf.write(str(conf))
 
@@ -292,6 +446,12 @@ class Installer:
 		return False
 
 	def mkinitcpio(self, *flags):
+		for plugin in plugins.values():
+			if hasattr(plugin, 'on_mkinitcpio'):
+				# Allow plugins to override the usage of mkinitcpio altogether.
+				if plugin.on_mkinitcpio(self):
+					return True
+
 		with open(f'{self.target}/etc/mkinitcpio.conf', 'w') as mkinit:
 			mkinit.write(f"MODULES=({' '.join(self.MODULES)})\n")
 			mkinit.write(f"BINARIES=({' '.join(self.BINARIES)})\n")
@@ -318,9 +478,14 @@ class Installer:
 			if partition.filesystem == 'btrfs':
 				if 'btrfs' not in self.MODULES:
 					self.MODULES.append('btrfs')
-				if '/usr/bin/btrfs-progs' not in self.BINARIES:
+				if '/usr/bin/btrfs' not in self.BINARIES:
 					self.BINARIES.append('/usr/bin/btrfs')
 
+			# There is not yet an fsck tool for NTFS. If it's being used for the root filesystem, the hook should be removed.
+			if partition.filesystem == 'ntfs3' and partition.mountpoint == self.target:
+				if 'fsck' in self.HOOKS:
+					self.HOOKS.remove('fsck')
+
 			if self.detect_encryption(partition):
 				if 'encrypt' not in self.HOOKS:
 					self.HOOKS.insert(self.HOOKS.index('filesystems'), 'encrypt')
@@ -344,9 +509,6 @@ class Installer:
 		self.pacstrap(self.base_packages)
 		self.helper_flags['base-strapped'] = True
 
-		with open(f"{self.target}/etc/fstab", "a") as fstab:
-			fstab.write("\ntmpfs /tmp tmpfs defaults,noatime,mode=1777 0 0\n")  # Redundant \n at the start? who knows?
-
 		# TODO: Support locale and timezone
 		# os.remove(f'{self.target}/etc/localtime')
 		# sys_command(f'/usr/bin/arch-chroot {self.target} ln -s /usr/share/zoneinfo/{localtime} /etc/localtime')
@@ -366,16 +528,49 @@ class Installer:
 			self.log(f"Running post-installation hook: {function}", level=logging.INFO)
 			function(self)
 
+		for plugin in plugins.values():
+			if hasattr(plugin, 'on_install'):
+				plugin.on_install(self)
+
 		return True
 
+	def setup_swap(self, kind='zram'):
+		if kind == 'zram':
+			self.log(f"Setting up swap on zram")
+			self.pacstrap('zram-generator')
+
+			# We could use the default example below, but maybe not the best idea: https://github.com/archlinux/archinstall/pull/678#issuecomment-962124813
+			# zram_example_location = '/usr/share/doc/zram-generator/zram-generator.conf.example'
+			# shutil.copy2(f"{self.target}{zram_example_location}", f"{self.target}/usr/lib/systemd/zram-generator.conf")
+			with open(f"{self.target}/etc/systemd/zram-generator.conf", "w") as zram_conf:
+				zram_conf.write("[zram0]\n")
+
+			if self.enable_service('systemd-zram-setup@zram0.service'):
+				return True
+		else:
+			raise ValueError(f"Archinstall currently only supports setting up swap on zram")
+
 	def add_bootloader(self, bootloader='systemd-bootctl'):
+		for plugin in plugins.values():
+			if hasattr(plugin, 'on_add_bootloader'):
+				# Allow plugins to override the boot-loader handling.
+				# This allows for bot configuring and installing bootloaders.
+				if plugin.on_add_bootloader(self):
+					return True
+
 		boot_partition = None
 		root_partition = None
+		root_partition_fs = None
 		for partition in self.partitions:
 			if partition.mountpoint == self.target + '/boot':
 				boot_partition = partition
 			elif partition.mountpoint == self.target:
 				root_partition = partition
+				root_partition_fs = partition.filesystem
+				root_fs_type = get_mount_fs_type(root_partition_fs)
+
+		if boot_partition is None and root_partition is None:
+			raise ValueError(f"Could not detect root (/) or boot (/boot) in {self.target} based on: {self.partitions}")
 
 		self.log(f'Adding bootloader {bootloader} to {boot_partition if boot_partition else root_partition}', level=logging.INFO)
 
@@ -393,6 +588,10 @@ class Installer:
 				# Fallback, try creating the boot loader without touching the EFI variables
 				SysCommand(f'/usr/bin/arch-chroot {self.target} bootctl --no-variables --path=/boot install')
 
+			# Ensure that the /boot/loader directory exists before we try to create files in it
+			if not os.path.exists(f'{self.target}/boot/loader'):
+				os.makedirs(f'{self.target}/boot/loader')
+
 			# Modify or create a loader.conf
 			if os.path.isfile(f'{self.target}/boot/loader/loader.conf'):
 				with open(f'{self.target}/boot/loader/loader.conf', 'r') as loader:
@@ -406,66 +605,124 @@ class Installer:
 			with open(f'{self.target}/boot/loader/loader.conf', 'w') as loader:
 				for line in loader_data:
 					if line[:8] == 'default ':
-						loader.write(f'default {self.init_time}\n')
+						loader.write(f'default {self.init_time}_{self.kernels[0]}\n')
 					elif line[:8] == '#timeout' and 'timeout 5' not in loader_data:
 						# We add in the default timeout to support dual-boot
 						loader.write(f"{line[1:]}\n")
 					else:
 						loader.write(f"{line}\n")
 
-			# For some reason, blkid and /dev/disk/by-uuid are not getting along well.
-			# And blkid is wrong in terms of LUKS.
-			# UUID = sys_command('blkid -s PARTUUID -o value {drive}{partition_2}'.format(**args)).decode('UTF-8').strip()
-			# Setup the loader entry
-			with open(f'{self.target}/boot/loader/entries/{self.init_time}.conf', 'w') as entry:
-				entry.write('# Created by: archinstall\n')
-				entry.write(f'# Created on: {self.init_time}\n')
-				entry.write('title Arch Linux\n')
-				entry.write('linux /vmlinuz-linux\n')
+			# Ensure that the /boot/loader/entries directory exists before we try to create files in it
+			if not os.path.exists(f'{self.target}/boot/loader/entries'):
+				os.makedirs(f'{self.target}/boot/loader/entries')
+
+			for kernel in self.kernels:
+				# Setup the loader entry
+				with open(f'{self.target}/boot/loader/entries/{self.init_time}_{kernel}.conf', 'w') as entry:
+					entry.write('# Created by: archinstall\n')
+					entry.write(f'# Created on: {self.init_time}\n')
+					entry.write(f'title Arch Linux ({kernel})\n')
+					entry.write(f"linux /vmlinuz-{kernel}\n")
+					if not is_vm():
+						vendor = cpu_vendor()
+						if vendor == "AuthenticAMD":
+							entry.write("initrd /amd-ucode.img\n")
+						elif vendor == "GenuineIntel":
+							entry.write("initrd /intel-ucode.img\n")
+						else:
+							self.log("unknow cpu vendor, not adding ucode to systemd-boot config")
+					entry.write(f"initrd /initramfs-{kernel}.img\n")
+					# blkid doesn't trigger on loopback devices really well,
+					# so we'll use the old manual method until we get that sorted out.
+
+					if real_device := self.detect_encryption(root_partition):
+						# TODO: We need to detect if the encrypted device is a whole disk encryption,
+						#       or simply a partition encryption. Right now we assume it's a partition (and we always have)
+						log(f"Identifying root partition by PART-UUID on {real_device}: '{real_device.uuid}'.", level=logging.DEBUG)
+						entry.write(f'options cryptdevice=PARTUUID={real_device.uuid}:luksdev root=/dev/mapper/luksdev rw intel_pstate=no_hwp rootfstype={root_fs_type} {" ".join(self.KERNEL_PARAMS)}\n')
+					else:
+						log(f"Identifying root partition by PART-UUID on {root_partition}, looking for '{root_partition.uuid}'.", level=logging.DEBUG)
+						entry.write(f'options root=PARTUUID={root_partition.uuid} rw intel_pstate=no_hwp rootfstype={root_fs_type} {" ".join(self.KERNEL_PARAMS)}\n')
+
+					self.helper_flags['bootloader'] = bootloader
+
+		elif bootloader == "grub-install":
+			self.pacstrap('grub')  # no need?
+
+			if real_device := self.detect_encryption(root_partition):
+				root_uuid = SysCommand(f"blkid -s UUID -o value {real_device.path}").decode().rstrip()
+				_file = "/etc/default/grub"
+				add_to_CMDLINE_LINUX = f"sed -i 's/GRUB_CMDLINE_LINUX=\"\"/GRUB_CMDLINE_LINUX=\"cryptdevice=UUID={root_uuid}:cryptlvm rootfstype={root_fs_type}\"/'"
+				enable_CRYPTODISK = "sed -i 's/#GRUB_ENABLE_CRYPTODISK=y/GRUB_ENABLE_CRYPTODISK=y/'"
+
+				log(f"Using UUID {root_uuid} of {real_device} as encrypted root identifier.", level=logging.INFO)
+				SysCommand(f"/usr/bin/arch-chroot {self.target} {add_to_CMDLINE_LINUX} {_file}")
+				SysCommand(f"/usr/bin/arch-chroot {self.target} {enable_CRYPTODISK} {_file}")
+			else:
+				_file = "/etc/default/grub"
+				add_to_CMDLINE_LINUX = f"sed -i 's/GRUB_CMDLINE_LINUX=\"\"/GRUB_CMDLINE_LINUX=\"rootfstype={root_fs_type}\"/'"
+				SysCommand(f"/usr/bin/arch-chroot {self.target} {add_to_CMDLINE_LINUX} {_file}")
+
+			log(f"GRUB uses {boot_partition.path} as the boot partition.", level=logging.INFO)
+			if has_uefi():
+				self.pacstrap('efibootmgr') # TODO: Do we need? Yes, but remove from minimal_installation() instead?
+				if not (handle := SysCommand(f'/usr/bin/arch-chroot {self.target} grub-install --target=x86_64-efi --efi-directory=/boot --bootloader-id=GRUB')).exit_code == 0:
+					raise DiskError(f"Could not install GRUB to {self.target}/boot: {handle}")
+			else:
+				if not (handle := SysCommand(f'/usr/bin/arch-chroot {self.target} grub-install --target=i386-pc --recheck {boot_partition.parent}')).exit_code == 0:
+					raise DiskError(f"Could not install GRUB to {boot_partition.path}: {handle}")
+
+			if not (handle := SysCommand(f'/usr/bin/arch-chroot {self.target} grub-mkconfig -o /boot/grub/grub.cfg')).exit_code == 0:
+				raise DiskError(f"Could not configure GRUB: {handle}")
+
+			self.helper_flags['bootloader'] = True
+		elif bootloader == 'efistub':
+			self.pacstrap('efibootmgr')
+
+			if not has_uefi():
+				raise HardwareIncompatibilityError
+			# TODO: Ideally we would want to check if another config
+			# points towards the same disk and/or partition.
+			# And in which case we should do some clean up.
+
+			for kernel in self.kernels:
+				# Setup the firmware entry
+
+				label = f'Arch Linux ({kernel})'
+				loader = f"/vmlinuz-{kernel}"
+
+				kernel_parameters = []
+
 				if not is_vm():
 					vendor = cpu_vendor()
 					if vendor == "AuthenticAMD":
-						entry.write("initrd /amd-ucode.img\n")
+						kernel_parameters.append("initrd=\\amd-ucode.img")
 					elif vendor == "GenuineIntel":
-						entry.write("initrd /intel-ucode.img\n")
+						kernel_parameters.append("initrd=\\intel-ucode.img")
 					else:
-						self.log("unknow cpu vendor, not adding ucode to systemd-boot config")
-				entry.write('initrd /initramfs-linux.img\n')
+						self.log("unknow cpu vendor, not adding ucode to firmware boot entry")
+
+				kernel_parameters.append(f"initrd=\\initramfs-{kernel}.img")
+
 				# blkid doesn't trigger on loopback devices really well,
 				# so we'll use the old manual method until we get that sorted out.
-
 				if real_device := self.detect_encryption(root_partition):
 					# TODO: We need to detect if the encrypted device is a whole disk encryption,
 					#       or simply a partition encryption. Right now we assume it's a partition (and we always have)
 					log(f"Identifying root partition by PART-UUID on {real_device}: '{real_device.uuid}'.", level=logging.DEBUG)
-					entry.write(f'options cryptdevice=PARTUUID={real_device.uuid}:luksdev root=/dev/mapper/luksdev rw intel_pstate=no_hwp {" ".join(self.KERNEL_PARAMS)}\n')
+					kernel_parameters.append(f'cryptdevice=PARTUUID={real_device.uuid}:luksdev root=/dev/mapper/luksdev rw intel_pstate=no_hwp rootfstype={root_fs_type} {" ".join(self.KERNEL_PARAMS)}')
 				else:
 					log(f"Identifying root partition by PART-UUID on {root_partition}, looking for '{root_partition.uuid}'.", level=logging.DEBUG)
-					entry.write(f'options root=PARTUUID={root_partition.uuid} rw intel_pstate=no_hwp {" ".join(self.KERNEL_PARAMS)}\n')
+					kernel_parameters.append(f'root=PARTUUID={root_partition.uuid} rw intel_pstate=no_hwp rootfstype={root_fs_type} {" ".join(self.KERNEL_PARAMS)}')
 
-				self.helper_flags['bootloader'] = bootloader
-				return True
+				SysCommand(f'efibootmgr --disk {boot_partition.path[:-1]} --part {boot_partition.path[-1]} --create --label "{label}" --loader {loader} --unicode \'{" ".join(kernel_parameters)}\' --verbose')
 
-		elif bootloader == "grub-install":
-			self.pacstrap('grub')
-
-			if has_uefi():
-				self.pacstrap('efibootmgr')
-				o = b''.join(SysCommand(f'/usr/bin/arch-chroot {self.target} grub-install --target=x86_64-efi --efi-directory=/boot --bootloader-id=GRUB'))
-				SysCommand('/usr/bin/arch-chroot /mnt grub-mkconfig -o /boot/grub/grub.cfg')
-				self.helper_flags['bootloader'] = True
-				return True
-			else:
-				root_device = subprocess.check_output(f'basename "$(readlink -f /sys/class/block/{root_partition.path.replace("/dev/", "")}/..)"', shell=True).decode().strip()
-				if root_device == "block":
-					root_device = f"{root_partition.path}"
-				o = b''.join(SysCommand(f'/usr/bin/arch-chroot {self.target} grub-install --target=i386-pc /dev/{root_device}'))
-				SysCommand('/usr/bin/arch-chroot /mnt grub-mkconfig -o /boot/grub/grub.cfg')
-				self.helper_flags['bootloader'] = True
-				return True
+			self.helper_flags['bootloader'] = bootloader
 		else:
 			raise RequirementError(f"Unknown (or not yet implemented) bootloader requested: {bootloader}")
 
+		return True
+
 	def add_additional_packages(self, *packages):
 		return self.pacstrap(*packages)
 
@@ -487,14 +744,25 @@ class Installer:
 	def user_create(self, user: str, password=None, groups=None, sudo=False):
 		if groups is None:
 			groups = []
-		self.log(f'Creating user {user}', level=logging.INFO)
-		o = b''.join(SysCommand(f'/usr/bin/arch-chroot {self.target} useradd -m -G wheel {user}'))
+
+		# This plugin hook allows for the plugin to handle the creation of the user.
+		# Password and Group management is still handled by user_create()
+		handled_by_plugin = False
+		for plugin in plugins.values():
+			if hasattr(plugin, 'on_user_create'):
+				if result := plugin.on_user_create(user):
+					handled_by_plugin = result
+
+		if not handled_by_plugin:
+			self.log(f'Creating user {user}', level=logging.INFO)
+			SysCommand(f'/usr/bin/arch-chroot {self.target} useradd -m -G wheel {user}')
+
 		if password:
 			self.user_set_pw(user, password)
 
 		if groups:
 			for group in groups:
-				o = b''.join(SysCommand(f'/usr/bin/arch-chroot {self.target} gpasswd -a {user} {group}'))
+				SysCommand(f'/usr/bin/arch-chroot {self.target} gpasswd -a {user} {group}')
 
 		if sudo and self.enable_sudo(user):
 			self.helper_flags['user'] = True
@@ -506,16 +774,21 @@ class Installer:
 			# This means the root account isn't locked/disabled with * in /etc/passwd
 			self.helper_flags['user'] = True
 
-		o = b''.join(SysCommand(f"/usr/bin/arch-chroot {self.target} sh -c \"echo '{user}:{password}' | chpasswd\""))
-		pass
+		SysCommand(f"/usr/bin/arch-chroot {self.target} sh -c \"echo '{user}:{password}' | chpasswd\"")
 
 	def user_set_shell(self, user, shell):
 		self.log(f'Setting shell for {user} to {shell}', level=logging.INFO)
 
-		o = b''.join(SysCommand(f"/usr/bin/arch-chroot {self.target} sh -c \"chsh -s {shell} {user}\""))
-		pass
+		SysCommand(f"/usr/bin/arch-chroot {self.target} sh -c \"chsh -s {shell} {user}\"")
+
+	def chown(self, owner, path, options=[]):
+		return SysCommand(f"/usr/bin/arch-chroot {self.target} sh -c 'chown {' '.join(options)} {owner} {path}")
+
+	def create_file(self, filename, owner=None):
+		return InstallationFile(self, filename, owner)
 
 	def set_keyboard_language(self, language: str) -> bool:
+		log(f"Setting keyboard language to {language}", level=logging.INFO)
 		if len(language.strip()):
 			if not verify_keyboard_layout(language):
 				self.log(f"Invalid keyboard language specified: {language}", fg="red", level=logging.ERROR)
@@ -524,7 +797,6 @@ class Installer:
 			# In accordance with https://github.com/archlinux/archinstall/issues/107#issuecomment-841701968
 			# Setting an empty keymap first, allows the subsequent call to set layout for both console and x11.
 			from .systemd import Boot
-
 			with Boot(self) as session:
 				session.SysCommand(["localectl", "set-keymap", '""'])
 
@@ -538,6 +810,7 @@ class Installer:
 		return True
 
 	def set_x11_keyboard_language(self, language: str) -> bool:
+		log(f"Setting x11 keyboard language to {language}", level=logging.INFO)
 		"""
 		A fallback function to set x11 layout specifically and separately from console layout.
 		This isn't strictly necessary since .set_keyboard_language() does this as well.
@@ -548,7 +821,6 @@ class Installer:
 				return False
 
 			from .systemd import Boot
-
 			with Boot(self) as session:
 				session.SysCommand(["localectl", "set-x11-keymap", '""'])
 
