@@ -7,18 +7,16 @@ import select  # Used for char by char polling of sys.stdin
 import shutil
 import signal
 import sys
-import termios
 import time
-import tty
 
-from .exceptions import *
-from .general import SysCommand
-from .hardware import AVAILABLE_GFX_DRIVERS, has_uefi
+from .disk import BlockDevice, valid_fs_type, suggest_single_disk_layout, suggest_multi_disk_layout, valid_parted_position
+from .exceptions import RequirementError, UserError, DiskError
+from .hardware import AVAILABLE_GFX_DRIVERS, has_uefi, has_amd_graphics, has_intel_graphics, has_nvidia_graphics
 from .locale_helpers import list_keyboard_languages, verify_keyboard_layout, search_keyboard_layout
 from .networking import list_interfaces
 from .output import log
 from .profiles import Profile, list_profiles
-
+from .storage import storage
 
 # TODO: Some inconsistencies between the selection processes.
 #       Some return the keys from the options, some the values?
@@ -121,7 +119,7 @@ def print_large_list(options, padding=5, margin_bottom=0, separator=': '):
 
 def generic_multi_select(options, text="Select one or more of the options above (leave blank to continue): ", sort=True, default=None, allow_empty=False):
 	# Checking if the options are different from `list` or `dict` or if they are empty
-	if type(options) not in [list, dict]:
+	if type(options) not in [list, dict, type({}.keys()), type({}.values())]:
 		log(f" * Generic multi-select doesn't support ({type(options)}) as type of options * ", fg='red')
 		log(" * If problem persists, please create an issue on https://github.com/archlinux/archinstall/issues * ", fg='yellow')
 		raise RequirementError("generic_multi_select() requires list or dictionary as options.")
@@ -132,6 +130,8 @@ def generic_multi_select(options, text="Select one or more of the options above
 	# After passing the checks, function continues to work
 	if type(options) == dict:
 		options = list(options.values())
+	elif type(options) in (type({}.keys()), type({}.values())):
+		options = list(options)
 	if sort:
 		options = sorted(options)
 
@@ -188,8 +188,25 @@ def generic_multi_select(options, text="Select one or more of the options above
 		except RequirementError as e:
 			log(f" * {e} * ", fg='red')
 
+	sys.stdout.write('\n')
+	sys.stdout.flush()
 	return selected_options
 
+def select_encrypted_partitions(block_devices :dict, password :str) -> dict:
+	for device in block_devices:
+		for partition in block_devices[device]['partitions']:
+			if partition.get('mountpoint', None) != '/boot':
+				partition['encrypted'] = True
+				partition['!password'] = password
+
+				if partition['mountpoint'] != '/':
+					# Tell the upcoming steps to generate a key-file for non root mounts.
+					partition['generate-encryption-key-file'] = True
+
+	return block_devices
+
+	# TODO: Next version perhaps we can support mixed multiple encrypted partitions
+	# Users might want to single out a partition for non-encryption to share between dualboot etc.
 
 class MiniCurses:
 	def __init__(self, width, height):
@@ -269,6 +286,8 @@ class MiniCurses:
 
 	def get_keyboard_input(self, strip_rowbreaks=True, end='\n'):
 		assert end in ['\r', '\n', None]
+		import termios
+		import tty
 
 		poller = select.epoll()
 		response = ''
@@ -308,6 +327,10 @@ class MiniCurses:
 			return response
 
 
+def ask_for_swap(prompt='Would you like to use swap on zram? (Y/n): ', forced=False):
+	return True if input(prompt).strip(' ').lower() not in ('n', 'no') else False
+
+
 def ask_for_superuser_account(prompt='Username for required superuser with sudo privileges: ', forced=False):
 	while 1:
 		new_user = input(prompt).strip(' ')
@@ -361,22 +384,34 @@ def ask_for_a_timezone():
 			)
 
 
-def ask_for_bootloader() -> str:
-	bootloader = "systemd-bootctl"
-	if not has_uefi():
-		bootloader = "grub-install"
-	else:
-		bootloader_choice = input("Would you like to use GRUB as a bootloader instead of systemd-boot? [y/N] ").lower()
-		if bootloader_choice == "y":
-			bootloader = "grub-install"
+def ask_for_bootloader(advanced_options=False) -> str:
+	bootloader = "systemd-bootctl" if has_uefi() else "grub-install"
+	if has_uefi():
+		if not advanced_options:
+			bootloader_choice = input("Would you like to use GRUB as a bootloader instead of systemd-boot? [y/N] ").lower()
+			if bootloader_choice == "y":
+				bootloader = "grub-install"
+		else:
+			# We use the common names for the bootloader as the selection, and map it back to the expected values.
+			choices = ['systemd-boot', 'grub', 'efistub']
+			selection = generic_select(choices, f'Choose a bootloader or leave blank to use systemd-boot: ', options_output=True)
+			if selection != "":
+				if selection == 'systemd-boot':
+					bootloader = 'systemd-bootctl'
+				elif selection == 'grub':
+					bootloader = 'grub-install'
+				else:
+					bootloader = selection
+
 	return bootloader
 
 
-def ask_for_audio_selection():
-	audio = "pulseaudio"  # Default for most desktop environments
-	pipewire_choice = input("Would you like to install pipewire instead of pulseaudio as the default audio server? [Y/n] ").lower()
-	if pipewire_choice in ("y", ""):
-		audio = "pipewire"
+def ask_for_audio_selection(desktop=True):
+	audio = 'pipewire' if desktop else 'none'
+	choices = ['pipewire', 'pulseaudio'] if desktop else ['pipewire', 'pulseaudio', 'none']
+	selection = generic_select(choices, f'Choose an audio server or leave blank to use {audio}: ', options_output=True)
+	if selection != "":
+		audio = selection
 
 	return audio
 
@@ -387,13 +422,13 @@ def ask_to_configure_network():
 	# {MAC: Ifname}
 	interfaces = {
 		'ISO-CONFIG': 'Copy ISO network configuration to installation',
-		'NetworkManager': 'Use NetworkManager to control and manage your internet connection',
+		'NetworkManager': 'Use NetworkManager (necessary to configure internet graphically in GNOME and KDE)',
 		**list_interfaces()
 	}
 
 	nic = generic_select(interfaces, "Select one network interface to configure (leave blank to skip): ")
 	if nic and nic != 'Copy ISO network configuration to installation':
-		if nic == 'Use NetworkManager to control and manage your internet connection':
+		if nic == 'Use NetworkManager (necessary to configure internet graphically in GNOME and KDE)':
 			return {'nic': nic, 'NetworkManager': True}
 
 		# Current workaround:
@@ -459,7 +494,7 @@ def ask_for_disk_layout():
 	return next((key for key, val in options.items() if val == value), None)
 
 
-def ask_for_main_filesystem_format():
+def ask_for_main_filesystem_format(advanced_options=False):
 	options = {
 		'btrfs': 'btrfs',
 		'ext4': 'ext4',
@@ -467,6 +502,13 @@ def ask_for_main_filesystem_format():
 		'f2fs': 'f2fs'
 	}
 
+	advanced = {
+		'ntfs': 'ntfs'
+	}
+
+	if advanced_options:
+		options.update(advanced)
+
 	value = generic_select(options, "Select which filesystem your main partition should use (by number or name): ", allow_empty_input=False)
 	return next((key for key, val in options.items() if val == value), None)
 
@@ -535,6 +577,223 @@ def generic_select(options, input_text="Select one of the above by index or abso
 
 	return selected_option
 
+def partition_overlap(partitions :list, start :str, end :str) -> bool:
+	# TODO: Implement sanity check
+	return False
+
+def get_default_partition_layout(block_devices, advanced_options=False):
+	if len(block_devices) == 1:
+		return suggest_single_disk_layout(block_devices[0], advanced_options=advanced_options)
+	else:
+		return suggest_multi_disk_layout(block_devices, advanced_options=advanced_options)
+
+	# TODO: Implement sane generic layout for 2+ drives
+
+def manage_new_and_existing_partitions(block_device :BlockDevice) -> dict:
+	# if has_uefi():
+	# 	partition_type = 'gpt'
+	# else:
+	# 	partition_type = 'msdos'
+
+	# log(f"Selecting which partitions to re-use on {block_device}...", fg="yellow", level=logging.INFO)
+	# partitions = generic_multi_select(block_device.partitions.values(), "Select which partitions to re-use (the rest will be left alone): ", sort=True)
+	# partitions_to_wipe = generic_multi_select(partitions, "Which partitions do you wish to wipe (multiple can be selected): ", sort=True)
+
+	# mountpoints = {}
+	# struct = {
+	# 	"partitions" : []
+	# }
+	# for partition in partitions:
+	# 	mountpoint = input(f"Select a mountpoint (or skip) for {partition}: ").strip()
+
+	# 	part_struct = {}
+	# 	if mountpoint:
+	# 		part_struct['mountpoint'] = mountpoint
+	# 		if mountpoint == '/boot':
+	# 			part_struct['boot'] = True
+	# 			if has_uefi():
+	# 				part_struct['ESP'] = True
+	# 		elif mountpoint == '/' and
+	# 	if partition.uuid:
+	# 		part_struct['PARTUUID'] = partition.uuid
+	# 	if partition in partitions_to_wipe:
+	# 		part_struct['wipe'] = True
+
+	# 	struct['partitions'].append(part_struct)
+
+	# return struct
+
+	block_device_struct = {
+		"partitions" : [partition.__dump__() for partition in block_device.partitions.values()]
+	}
+	# Test code: [part.__dump__() for part in block_device.partitions.values()]
+	# TODO: Squeeze in BTRFS subvolumes here
+
+	while True:
+		modes = [
+			"Create a new partition",
+			f"Suggest partition layout for {block_device}",
+			"Delete a partition" if len(block_device_struct) else "",
+			"Clear/Delete all partitions" if len(block_device_struct) else "",
+			"Assign mount-point for a partition" if len(block_device_struct) else "",
+			"Mark/Unmark a partition to be formatted (wipes data)" if len(block_device_struct) else "",
+			"Mark/Unmark a partition as encrypted" if len(block_device_struct) else "",
+			"Mark/Unmark a partition as bootable (automatic for /boot)" if len(block_device_struct) else "",
+			"Set desired filesystem for a partition" if len(block_device_struct) else "",
+		]
+
+		# Print current partition layout:
+		if len(block_device_struct["partitions"]):
+			print('Current partition layout:')
+			for partition in block_device_struct["partitions"]:
+				print(partition)
+			print()
+
+		task = generic_select(modes,
+				input_text=f"Select what to do with {block_device} (leave blank when done): ")
+
+		if not task:
+			break
+
+		if task == 'Create a new partition':
+			# if partition_type == 'gpt':
+			# 	# https://www.gnu.org/software/parted/manual/html_node/mkpart.html
+			# 	# https://www.gnu.org/software/parted/manual/html_node/mklabel.html
+			# 	name = input("Enter a desired name for the partition: ").strip()
+
+			fstype = input("Enter a desired filesystem type for the partition: ").strip()
+
+			start = input(f"Enter the start sector (percentage or block number, default: {block_device.first_free_sector}): ").strip()
+			if not start.strip():
+				start = block_device.first_free_sector
+				end_suggested = block_device.first_end_sector
+			else:
+				end_suggested = '100%'
+			end = input(f"Enter the end sector of the partition (percentage or block number, ex: {end_suggested}): ").strip()
+			if not end.strip():
+				end = end_suggested
+
+			if valid_parted_position(start) and valid_parted_position(end) and valid_fs_type(fstype):
+				if partition_overlap(block_device_struct["partitions"], start, end):
+					log(f"This partition overlaps with other partitions on the drive! Ignoring this partition creation.", fg="red")
+					continue
+
+				block_device_struct["partitions"].append({
+					"type" : "primary", # Strictly only allowed under MSDOS, but GPT accepts it so it's "safe" to inject
+					"start" : start,
+					"size" : end,
+					"mountpoint" : None,
+					"wipe" : True,
+					"filesystem" : {
+						"format" : fstype
+					}
+				})
+			else:
+				log(f"Invalid start ({valid_parted_position(start)}), end ({valid_parted_position(end)}) or fstype ({valid_fs_type(fstype)}) for this partition. Ignoring this partition creation.", fg="red")
+				continue
+		elif task[:len("Suggest partition layout")] == "Suggest partition layout":
+			if len(block_device_struct["partitions"]):
+				if input(f"{block_device} contains queued partitions, this will remove those, are you sure? y/N: ").strip().lower() in ('', 'n'):
+					continue
+
+			block_device_struct.update(suggest_single_disk_layout(block_device)[block_device.path])
+		elif task is None:
+			return block_device_struct
+		else:
+			for index, partition in enumerate(block_device_struct["partitions"]):
+				print(f"{index}: Start: {partition['start']}, End: {partition['size']} ({partition['filesystem']['format']}{', mounting at: '+partition['mountpoint'] if partition['mountpoint'] else ''})")
+
+			if task == "Delete a partition":
+				if (partition := generic_select(block_device_struct["partitions"], 'Select which partition to delete: ', options_output=False)):
+					del(block_device_struct["partitions"][block_device_struct["partitions"].index(partition)])
+			elif task == "Clear/Delete all partitions":
+				block_device_struct["partitions"] = []
+			elif task == "Assign mount-point for a partition":
+				if (partition := generic_select(block_device_struct["partitions"], 'Select which partition to mount where: ', options_output=False)):
+					print(' * Partition mount-points are relative to inside the installation, the boot would be /boot as an example.')
+					mountpoint = input('Select where to mount partition (leave blank to remove mountpoint): ').strip()
+
+					if len(mountpoint):
+						block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['mountpoint'] = mountpoint
+						if mountpoint == '/boot':
+							log(f"Marked partition as bootable because mountpoint was set to /boot.", fg="yellow")
+							block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['boot'] = True
+					else:
+						del(block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['mountpoint'])
+
+			elif task == "Mark/Unmark a partition to be formatted (wipes data)":
+				if (partition := generic_select(block_device_struct["partitions"], 'Select which partition to mask for formatting: ', options_output=False)):
+					# If we mark a partition for formatting, but the format is CRYPTO LUKS, there's no point in formatting it really
+					# without asking the user which inner-filesystem they want to use. Since the flag 'encrypted' = True is already set,
+					# it's safe to change the filesystem for this partition.
+					if block_device_struct["partitions"][block_device_struct["partitions"].index(partition)].get('filesystem', {}).get('format', 'crypto_LUKS') == 'crypto_LUKS':
+						if not block_device_struct["partitions"][block_device_struct["partitions"].index(partition)].get('filesystem', None):
+							block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['filesystem'] = {}
+
+						while True:
+							fstype = input("Enter a desired filesystem type for the partition: ").strip()
+							if not valid_fs_type(fstype):
+								log(f"Desired filesystem {fstype} is not a valid filesystem.", level=logging.ERROR, fg="red")
+								continue
+							break
+
+						block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['filesystem']['format'] = fstype
+
+					# Negate the current wipe marking
+					block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['format'] = not block_device_struct["partitions"][block_device_struct["partitions"].index(partition)].get('format', False)
+
+			elif task == "Mark/Unmark a partition as encrypted":
+				if (partition := generic_select(block_device_struct["partitions"], 'Select which partition to mark as encrypted: ', options_output=False)):
+					# Negate the current encryption marking
+					block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['encrypted'] = not block_device_struct["partitions"][block_device_struct["partitions"].index(partition)].get('encrypted', False)
+
+			elif task == "Mark/Unmark a partition as bootable (automatic for /boot)":
+				if (partition := generic_select(block_device_struct["partitions"], 'Select which partition to mark as bootable: ', options_output=False)):
+					block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['boot'] = not block_device_struct["partitions"][block_device_struct["partitions"].index(partition)].get('boot', False)
+
+			elif task == "Set desired filesystem for a partition":
+				if not block_device_struct["partitions"]:
+					log("No partitions found. Create some partitions first", level=logging.WARNING, fg='yellow')
+					continue
+				elif (partition := generic_select(block_device_struct["partitions"], 'Select which partition to set a filesystem on: ', options_output=False)):
+					if not block_device_struct["partitions"][block_device_struct["partitions"].index(partition)].get('filesystem', None):
+						block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['filesystem'] = {}
+
+					while True:
+						fstype = input("Enter a desired filesystem type for the partition: ").strip()
+						if not valid_fs_type(fstype):
+							log(f"Desired filesystem {fstype} is not a valid filesystem.", level=logging.ERROR, fg="red")
+							continue
+						break
+
+					block_device_struct["partitions"][block_device_struct["partitions"].index(partition)]['filesystem']['format'] = fstype
+
+	return block_device_struct
+
+def select_individual_blockdevice_usage(block_devices :list):
+	result = {}
+
+	for device in block_devices:
+		layout = manage_new_and_existing_partitions(device)
+
+		result[device.path] = layout
+
+	return result
+
+
+def select_disk_layout(block_devices :list, advanced_options=False):
+	modes = [
+		"Wipe all selected drives and use a best-effort default partition layout",
+		"Select what to do with each individual drive (followed by partition usage)"
+	]
+
+	mode = generic_select(modes, input_text=f"Select what you wish to do with the selected block devices: ")
+
+	if mode == 'Wipe all selected drives and use a best-effort default partition layout':
+		return get_default_partition_layout(block_devices, advanced_options)
+	else:
+		return select_individual_blockdevice_usage(block_devices)
+
 
 def select_disk(dict_o_disks):
 	"""
@@ -575,11 +834,11 @@ def select_profile():
 
 	if len(shown_profiles) >= 1:
 		for index, profile in enumerate(shown_profiles):
-			print(f"{index}: {profile}")
+			description = Profile(None, profile).get_profile_description()
+			print(f"{index}: {profile}: {description}")
 
 		print(' -- The above list is a set of pre-programmed profiles. --')
 		print(' -- They might make it easier to install things like desktop environments. --')
-		print(' -- The desktop profile will let you select a DE/WM profile, e.g gnome, kde, sway --')
 		print(' -- (Leave blank and hit enter to skip this step and continue) --')
 
 		selected_profile = generic_select(actual_profiles_raw, 'Enter a pre-programmed profile name if you want to install one: ', options_output=False)
@@ -699,32 +958,23 @@ def select_driver(options=AVAILABLE_GFX_DRIVERS):
 	"""
 
 	drivers = sorted(list(options))
-	default_option = options["All open-source (default)"]
 
 	if drivers:
-		for line in SysCommand('/usr/bin/lspci'):
-			if b' vga ' in line.lower():
-				if b'nvidia' in line.lower():
-					print(' ** nvidia card detected, suggested driver: nvidia **')
-				elif b'amd' in line.lower():
-					print(' ** AMD card detected, suggested driver: AMD / ATI **')
+		arguments = storage.get('arguments', {})
+		if has_amd_graphics():
+			print('For the best compatibility with your AMD hardware, you may want to use either the all open-source or AMD / ATI options.')
+		if has_intel_graphics():
+			print('For the best compatibility with your Intel hardware, you may want to use either the all open-source or Intel options.')
+		if has_nvidia_graphics():
+			print('For the best compatibility with your Nvidia hardware, you may want to use the Nvidia proprietary driver.')
 
-		initial_option = generic_select(drivers, input_text="Select your graphics card driver: ")
+		if not arguments.get('gfx_driver', None):
+			arguments['gfx_driver'] = generic_select(drivers, input_text="Select a graphics driver or leave blank to install all open-source drivers: ")
 
-		if not initial_option:
-			return default_option
+		if arguments.get('gfx_driver', None) is None:
+			arguments['gfx_driver'] = "All open-source (default)"
 
-		selected_driver = options[initial_option]
-
-		if type(selected_driver) == dict:
-			driver_options = sorted(list(selected_driver))
-
-			driver_package_group = generic_select(driver_options, f'Which driver-type do you want for {initial_option}: ', allow_empty_input=False)
-			driver_package_group = selected_driver[driver_package_group]
-
-			return driver_package_group
-
-		return selected_driver
+		return options.get(arguments.get('gfx_driver'))
 
 	raise RequirementError("Selecting drivers require a least one profile to be given as an option.")
 
