@@ -4,6 +4,15 @@ import archinstall
 
 is_top_level_profile = False
 
+# "It is recommended also to install the gnome group, which contains applications required for the standard GNOME experience." - Arch Wiki
+__packages__ = [
+	"budgie-desktop",
+	"gnome",
+	"lightdm",
+	"lightdm-gtk-greeter",
+]
+
+
 def _prep_function(*args, **kwargs):
 	"""
 	Magic function called by the importing installer
@@ -20,15 +29,15 @@ def _prep_function(*args, **kwargs):
 		else:
 			print('Deprecated (??): xorg profile has no _prep_function() anymore')
 
+
 # Ensures that this code only gets executed if executed
 # through importlib.util.spec_from_file_location("budgie", "/somewhere/budgie.py")
 # or through conventional import budgie
 if __name__ == 'budgie':
 	# Install dependency profiles
-	installation.install_profile('xorg')
+	archinstall.storage['installation_session'].install_profile('xorg')
 
-	# Install the application budgie from the template under /applications/
-	budgie = archinstall.Application(installation, 'budgie')
-	budgie.install()
+	# Install the Budgie packages
+	archinstall.storage['installation_session'].add_additional_packages(__packages__)
 
-	installation.enable_service('lightdm') # Light Display Manager
+	archinstall.storage['installation_session'].enable_service('lightdm')  # Light Display Manager
