@@ -10,6 +10,9 @@ __packages__ = [
 	"pavucontrol",
 	"lightdm",
 	"lightdm-gtk-greeter",
+	"gvfs",
+	"network-manager-applet",
+	"xarchiver"
 ]
 
 
