@@ -1,6 +1,7 @@
 # A system with "xorg" installed
 
 import archinstall
+import logging
 
 is_top_level_profile = True
 
@@ -11,7 +12,6 @@ __packages__ = [
 	'xorg-server',
 	'xorg-xinit',
 	'nvidia-dkms',
-	'xorg-server',
 	*archinstall.lib.hardware.__packages__,
 ]
 
@@ -37,13 +37,16 @@ def _prep_function(*args, **kwargs):
 # or through conventional import xorg
 if __name__ == 'xorg':
 	try:
-		if "nvidia" in archinstall.storage.get("gfx_driver_packages", None):
+		if "nvidia" in archinstall.storage.get("gfx_driver_packages", []):
 			if "linux-zen" in archinstall.storage['installation_session'].base_packages or "linux-lts" in archinstall.storage['installation_session'].base_packages:
+				for kernel in archinstall.storage['installation_session'].kernels:
+					archinstall.storage['installation_session'].add_additional_packages(f"{kernel}-headers") # Fixes https://github.com/archlinux/archinstall/issues/585
 				archinstall.storage['installation_session'].add_additional_packages("dkms")  # I've had kernel regen fail if it wasn't installed before nvidia-dkms
 				archinstall.storage['installation_session'].add_additional_packages("xorg-server xorg-xinit nvidia-dkms")
 			else:
-				archinstall.storage['installation_session'].add_additional_packages(f"xorg-server xorg-xinit {' '.join(archinstall.storage.get('gfx_driver_packages', None))}")
+				archinstall.storage['installation_session'].add_additional_packages(f"xorg-server xorg-xinit {' '.join(archinstall.storage.get('gfx_driver_packages', []))}")
 		else:
-			archinstall.storage['installation_session'].add_additional_packages(f"xorg-server xorg-xinit {' '.join(archinstall.storage.get('gfx_driver_packages', None))}")
-	except:
+			archinstall.storage['installation_session'].add_additional_packages(f"xorg-server xorg-xinit {' '.join(archinstall.storage.get('gfx_driver_packages', []))}")
+	except Exception as err:
+		archinstall.log(f"Could not handle nvidia and linuz-zen specific situations during xorg installation: {err}", level=logging.WARNING, fg="yellow")
 		archinstall.storage['installation_session'].add_additional_packages("xorg-server xorg-xinit")  # Prep didn't run, so there's no driver to install
